# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonObjectSchema
from xcube.webapi.common.schemas import STRING_SCHEMA

CONFIGURATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=STRING_SCHEMA,
    ),
    additional_properties=False,
)

VIEWER_SCHEMA = JsonObjectSchema(
    properties=dict(
        Configuration=CONFIGURATION_SCHEMA,
    ),
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        Viewer=VIEWER_SCHEMA,
    )
)



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


import fnmatch
import itertools
import os
import os.path
import warnings
from functools import cached_property
from typing import Any, Optional, Callable, Union
from collections.abc import Collection, Mapping

import numpy as np
import pandas as pd
import pyproj
import xarray as xr

from xcube.constants import LOG
from xcube.core.mldataset import BaseMultiLevelDataset
from xcube.core.mldataset import IdentityMultiLevelDataset
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.mldataset import augment_ml_dataset
from xcube.core.mldataset import open_ml_dataset_from_python_code
from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataStoreConfig
from xcube.core.store import DataStorePool
from xcube.core.store import DatasetDescriptor
from xcube.core.store import MULTI_LEVEL_DATASET_TYPE
from xcube.core.tile import get_var_cmap_params
from xcube.core.tile import get_var_valid_range
from xcube.core.tile import DEFAULT_CMAP_NAME
from xcube.core.tile import DEFAULT_CMAP_NORM
from xcube.core.tile import DEFAULT_VALUE_RANGE
from xcube.core.varexpr import split_var_assignment
from xcube.server.api import Context, ApiError
from xcube.server.api import ServerConfig
from xcube.server.config import is_absolute_path
from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.cache import parse_mem_size
from xcube.util.cmaps import ColormapRegistry
from xcube.util.cmaps import load_custom_colormap
from xcube.util.cmaps import create_colormap_from_config
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.places import PlacesContext

COMPUTE_DATASET = "compute_dataset"
COMPUTE_VARIABLES = "compute_variables"

# We use tilde, because it is not a reserved URI characters
STORE_DS_ID_SEPARATOR = "~"
DATA_STORE_IDS_WARNING = ["stac"]
FS_TYPE_TO_PROTOCOL = {
    "local": "file",
    "obs": "s3",
    "file": "file",
    "s3": "s3",
    "memory": "memory",
}
NON_MEMORY_FILE_SYSTEMS = ["local", "obs", "file", "s3"]

ALL_PLACES = "all"

MultiLevelDatasetOpener = Callable[["DatasetsContext", ServerConfig], MultiLevelDataset]

DatasetConfig = Mapping[str, Any]


class DatasetsContext(ResourcesContext):
    def __init__(self, server_ctx: Context, ml_dataset_openers=None):
        super().__init__(server_ctx)
        # noinspection PyTypeChecker
        self._places_ctx: PlacesContext = server_ctx.get_api_ctx("places")
        assert isinstance(self._places_ctx, PlacesContext)
        self._ml_dataset_openers = ml_dataset_openers
        # cache for all dataset configs
        # contains tuples of form (MultiLevelDataset, dataset_config)
        self._dataset_cache = dict()
        self._data_store_pool, self._dataset_configs = self._process_dataset_configs(
            self.config, self.base_dir
        )
        self._cm_styles, self._colormap_registry = self._get_cm_styles()

    def on_dispose(self):
        with self.rlock:
            # Close all datasets
            for ml_dataset, _ in self._dataset_cache.values():
                # noinspection PyBroadException
                try:
                    ml_dataset.close()
                except Exception:
                    pass
            # Clear all caches
            if self._dataset_cache:
                self._dataset_cache.clear()
            if self._data_store_pool:
                self._data_store_pool.remove_all_store_configs()
            self._dataset_configs = None

    @property
    def places_ctx(self):
        return self._places_ctx

    @property
    def dataset_cache(self) -> dict[str, tuple[MultiLevelDataset, DatasetConfig]]:
        return self._dataset_cache

    @cached_property
    def access_control(self) -> dict[str, Any]:
        return self.config.get("AccessControl", {})

    @cached_property
    def required_scopes(self) -> list[str]:
        return self.access_control.get("RequiredScopes", [])

    @property
    def colormap_registry(self) -> ColormapRegistry:
        return self._colormap_registry

    def get_required_dataset_scopes(self, dataset_config: DatasetConfig) -> set[str]:
        return self._get_required_scopes(
            dataset_config, "read:dataset", "Dataset", dataset_config["Identifier"]
        )

    def get_required_variable_scopes(
        self, dataset_config: DatasetConfig, var_name: str
    ) -> set[str]:
        return self._get_required_scopes(
            dataset_config, "read:variable", "Variable", var_name
        )

    def _get_required_scopes(
        self,
        dataset_config: DatasetConfig,
        base_scope: str,
        value_name: str,
        value: str,
    ) -> set[str]:
        required_global_scopes = set(self.required_scopes)
        required_dataset_scopes = set(
            dataset_config.get("AccessControl", {}).get("RequiredScopes", [])
        )
        if not required_global_scopes and not required_dataset_scopes:
            return required_global_scopes
        required_dataset_scopes = required_global_scopes.union(required_dataset_scopes)
        base_scope_prefix = base_scope + ":"
        required_dataset_scopes = {
            scope
            for scope in required_dataset_scopes
            if scope == base_scope or scope.startswith(base_scope_prefix)
        }
        pattern_scope = base_scope_prefix + "{" + value_name + "}"
        if pattern_scope in required_dataset_scopes:
            # Replace "{base_scope}:{value_name}" by "{base_scope}:{value}"
            required_dataset_scopes.remove(pattern_scope)
            required_dataset_scopes.add(base_scope_prefix + value)
        return required_dataset_scopes

    def get_ml_dataset(self, ds_id: str) -> MultiLevelDataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        return ml_dataset

    def set_ml_dataset(self, ml_dataset: MultiLevelDataset):
        self._set_dataset_entry(
            (ml_dataset, dict(Identifier=ml_dataset.ds_id, Hidden=True))
        )

    def add_dataset(
        self,
        dataset: Union[xr.Dataset, MultiLevelDataset],
        ds_id: Optional[str] = None,
        title: Optional[str] = None,
        style: Optional[str] = None,
        color_mappings: dict[str, dict[str, Any]] = None,
    ) -> str:
        """Add a dataset to the configuration"""
        assert_instance(dataset, (xr.Dataset, MultiLevelDataset), "dataset")
        if isinstance(dataset, xr.Dataset):
            ml_dataset = BaseMultiLevelDataset(dataset, ds_id=ds_id)
        else:
            ml_dataset = dataset
            if not ds_id:
                _ds_id = ml_dataset.ds_id
                # TODO: Fixes https://github.com/xcube-dev/xcube/issues/1007, but
                #   this does not solve the likely root cause
                #   in xcube.core.mldataset.fs.FsMultiLevelDataset, see ctor.
                if "/" in _ds_id:
                    ds_id = _ds_id.replace("/", "-")
            if ds_id:
                ml_dataset = IdentityMultiLevelDataset(dataset, ds_id=ds_id)
            dataset = ml_dataset.base_dataset
        ds_id = ml_dataset.ds_id
        dataset_configs = self.get_dataset_configs()
        for index, dataset_config in enumerate(dataset_configs):
            if dataset_config["Identifier"] == ds_id:
                del dataset_configs[index]
                break
        dataset_config = dict(
            Identifier=ds_id, Title=title or dataset.attrs.get("title", ds_id)
        )
        if style is not None:
            dataset_config.update(dict(Style=style))
        if color_mappings is not None:
            style = style or ds_id
            dataset_config.update(dict(Style=style))
            self._cm_styles[style] = color_mappings
        self._dataset_cache[ds_id] = ml_dataset, dataset_config
        self._dataset_configs.append(dataset_config)
        return ds_id

    def remove_dataset(self, ds_id: str):
        assert_instance(ds_id, str, "ds_id")
        assert_given(ds_id, "ds_id")
        if ds_id in self._dataset_cache:
            del self._dataset_cache[ds_id]
        self._dataset_configs = [
            dc for dc in self._dataset_configs if dc["Identifier"] != ds_id
        ]

    def add_ml_dataset(
        self,
        ml_dataset: MultiLevelDataset,
        ds_id: Optional[str] = None,
        title: Optional[str] = None,
    ):
        if ds_id:
            ml_dataset.ds_id = ds_id
        else:
            ds_id = ml_dataset.ds_id
        self._set_dataset_entry((ml_dataset, dict(Identifier=ds_id, Title=title)))

    def get_dataset(
        self, ds_id: str, expected_var_names: Collection[str] = None
    ) -> xr.Dataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        dataset = ml_dataset.base_dataset
        if expected_var_names:
            for var_name in expected_var_names:
                if var_name not in dataset:
                    raise ApiError.NotFound(
                        f'Variable "{var_name}" not found' f' in dataset "{ds_id}"'
                    )
        return dataset

    def get_time_series_dataset(self, ds_id: str, var_name: str = None) -> xr.Dataset:
        dataset_config = self.get_dataset_config(ds_id)
        ts_ds_name = dataset_config.get("TimeSeriesDataset", ds_id)
        try:
            # Try to get more efficient, time-chunked dataset
            return self.get_dataset(
                ts_ds_name, expected_var_names=[var_name] if var_name else None
            )
        except ApiError.NotFound:
            # This happens, if the dataset pointed to by 'TimeSeriesDataset'
            # does not contain the variable given by var_name.
            return self.get_dataset(
                ds_id, expected_var_names=[var_name] if var_name else None
            )

    def get_variable_for_z(
        self, ds_id: str, var_name: str, z_index: int
    ) -> xr.DataArray:
        ml_dataset = self.get_ml_dataset(ds_id)
        index = ml_dataset.num_levels - 1 - z_index
        if index < 0 or index >= ml_dataset.num_levels:
            raise ApiError.NotFound(
                f'Variable "{var_name}" has no z-index {z_index}'
                f' in dataset "{ds_id}"'
            )
        dataset = ml_dataset.get_dataset(index)
        if var_name not in dataset:
            raise ApiError.NotFound(
                f'Variable "{var_name}" not found' f' in dataset "{ds_id}"'
            )
        return dataset[var_name]

    @classmethod
    def _maybe_assign_store_instance_ids(
        cls,
        dataset_configs: list[dict[str, Any]],
        data_store_pool: DataStorePool,
        base_dir: str,
    ) -> None:
        assignable_dataset_configs = [
            dc
            for dc in dataset_configs
            if "StoreInstanceId" not in dc
            and dc.get("FileSystem", "file") in NON_MEMORY_FILE_SYSTEMS
        ]
        # split into sub-lists according to file system and
        # non-root store params
        config_lists = []
        for config in assignable_dataset_configs:
            store_params = cls._get_other_store_params_than_root(config)
            file_system = config.get("FileSystem", "file")
            appended = False
            for config_list in config_lists:
                if config_list[0] == file_system and config_list[1] == store_params:
                    config_list[2].append(config)
                    appended = True
                    break
            if not appended:
                config_lists.append((file_system, store_params, [config]))

        for file_system, store_params, config_list in config_lists:
            # Retrieve paths per configuration
            paths = [dc["Path"] for dc in config_list]
            # Determine common prefixes of paths (and call them roots)
            roots = _get_roots(paths)
            for root in roots:
                fs_protocol = FS_TYPE_TO_PROTOCOL.get(file_system, file_system)
                store_params_with_root = store_params.copy()
                if fs_protocol == "file" and not is_absolute_path(root):
                    store_params_with_root["root"] = f"{base_dir}/{root}"
                else:
                    store_params_with_root["root"] = root
                # See if there already is a store with this configuration
                data_store_config = DataStoreConfig(
                    store_id=fs_protocol, store_params=store_params_with_root
                )
                store_instance_id = data_store_pool.get_store_instance_id(
                    data_store_config
                )
                if not store_instance_id:
                    # Create new store with new unique store instance id
                    counter = 1
                    while data_store_pool.has_store_instance(
                        f"{fs_protocol}_{counter}"
                    ):
                        counter += 1
                    store_instance_id = f"{fs_protocol}_{counter}"
                    data_store_pool.add_store_config(
                        store_instance_id, data_store_config
                    )
                for config in config_list:
                    path = config["Path"]
                    if path.startswith(root) and config.get("StoreInstanceId") is None:
                        config["StoreInstanceId"] = store_instance_id
                        new_path = path[len(root) :]
                        while new_path.startswith("/") or new_path.startswith("\\"):
                            new_path = new_path[1:]
                        config["Path"] = new_path

    @classmethod
    def _get_other_store_params_than_root(cls, dataset_config: DatasetConfig) -> dict:
        file_system = FS_TYPE_TO_PROTOCOL.get(dataset_config.get("FileSystem", "file"))
        if file_system != "s3":
            return {}
        storage_options = dict()
        if "Anonymous" in dataset_config:
            storage_options["anon"] = dataset_config["Anonymous"]
        client_kwargs = dict()
        if "Endpoint" in dataset_config:
            client_kwargs["endpoint_url"] = dataset_config["Endpoint"]
        if "Region" in dataset_config:
            client_kwargs["region_name"] = dataset_config["Region"]
        storage_options["client_kwargs"] = client_kwargs
        store_params = dict(storage_options=storage_options)
        return store_params

    @classmethod
    def get_dataset_configs_from_stores(
        cls, data_store_pool: DataStorePool
    ) -> list[DatasetConfig]:
        all_dataset_configs: list[DatasetConfig] = []
        for store_instance_id in data_store_pool.store_instance_ids:
            LOG.info(f"Scanning store {store_instance_id!r}")
            data_store_config = data_store_pool.get_store_config(store_instance_id)

            # Note by forman: This iterator chaining is inefficient.
            # Preferably, we should offer
            #
            # store_dataset_ids = data_store.get_data_ids(
            #     data_type=(DATASET_TYPE, MULTI_LEVEL_DATASET_TYPE)
            # )
            #

            store_dataset_configs: list[ServerConfig] = data_store_config.user_data
            if store_dataset_configs:
                for store_dataset_config in store_dataset_configs:
                    dataset_id_pattern = store_dataset_config.get("Path", "*")
                    if _is_wildard(dataset_id_pattern):
                        if store_instance_id in DATA_STORE_IDS_WARNING:
                            warnings.warn(
                                f"The data store with ID '{store_instance_id}' has "
                                "many data IDs. Using wildcard patterns to select "
                                "datasets may cause a long setup time of the server."
                            )
                        data_store = data_store_pool.get_store(store_instance_id)
                        store_dataset_ids = itertools.chain(
                            data_store.get_data_ids(data_type=DATASET_TYPE),
                            data_store.get_data_ids(data_type=MULTI_LEVEL_DATASET_TYPE),
                        )
                        for store_dataset_id in store_dataset_ids:
                            if fnmatch.fnmatch(store_dataset_id, dataset_id_pattern):
                                all_dataset_configs.append(
                                    _get_selected_dataset_config(
                                        store_dataset_id,
                                        store_instance_id,
                                        store_dataset_config,
                                    )
                                )
                    else:
                        all_dataset_configs.append(
                            _get_selected_dataset_config(
                                store_dataset_config["Path"],
                                store_instance_id,
                                store_dataset_config,
                            )
                        )

        # # Just for testing:
        # debug_file = 'all_dataset_configs.json'
        # with open(debug_file, 'w') as stream:
        #     json.dump(all_dataset_configs, stream)
        #     LOG.debug(f'Wrote file {debug_file!r}')

        return all_dataset_configs

    def new_dataset_metadata(
        self, store_instance_id: str, dataset_id: str
    ) -> Optional[DatasetDescriptor]:
        data_store = self._data_store_pool.get_store(store_instance_id)
        dataset_metadata = data_store.describe_data(dataset_id, data_type="dataset")
        if dataset_metadata.crs is not None:
            crs = pyproj.CRS.from_string(dataset_metadata.crs)
            if not crs.is_geographic:
                LOG.warning(
                    f"Ignoring dataset {dataset_id!r} from"
                    f" store instance {store_instance_id!r}"
                    f" because it uses a non-geographic CRS"
                )
                return None
        # noinspection PyTypeChecker
        return dataset_metadata

    def get_dataset_config(self, ds_id: str) -> DatasetConfig:
        dataset_configs = self.get_dataset_configs()
        dataset_config = next(
            (dsd for dsd in dataset_configs if dsd["Identifier"] == ds_id), None
        )
        if dataset_config is None:
            raise ApiError.NotFound(f'Dataset "{ds_id}" not found')
        return dataset_config

    def get_dataset_configs(self) -> list[DatasetConfig]:
        assert self._dataset_configs is not None
        return self._dataset_configs

    def get_data_store_pool(self) -> DataStorePool:
        assert self._data_store_pool is not None
        return self._data_store_pool

    @classmethod
    def _process_dataset_configs(
        cls, config: ServerConfig, base_dir: str
    ) -> tuple[DataStorePool, list[dict[str, Any]]]:
        data_store_configs = config.get("DataStores", [])
        dataset_configs = config.get("Datasets", [])

        data_store_pool = DataStorePool()
        for data_store_config_dict in data_store_configs:
            store_instance_id = data_store_config_dict.get("Identifier")
            store_id = data_store_config_dict.get("StoreId")
            store_params = data_store_config_dict.get("StoreParams", {})
            store_dataset_configs = data_store_config_dict.get("Datasets")
            store_config = DataStoreConfig(
                store_id, store_params=store_params, user_data=store_dataset_configs
            )
            data_store_pool.add_store_config(store_instance_id, store_config)
        dataset_configs = dataset_configs + cls.get_dataset_configs_from_stores(
            data_store_pool
        )
        # Allow dataset_configs to be writable, because
        # _maybe_assign_store_instance_ids() will change
        # entries:
        dataset_configs = [dict(c) for c in dataset_configs]
        cls._maybe_assign_store_instance_ids(dataset_configs, data_store_pool, base_dir)
        return data_store_pool, dataset_configs

    def get_rgb_color_mapping(
        self, ds_id: str, norm_range: tuple[float, float] = (0.0, 1.0)
    ) -> tuple[list[Optional[str]], list[tuple[float, float]]]:
        var_names = [None, None, None]
        norm_ranges = [norm_range, norm_range, norm_range]
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            rgb_mapping = color_mappings.get("rgb")
            if rgb_mapping:
                components = "Red", "Green", "Blue"
                for i in range(3):
                    component = components[i]
                    component_config = rgb_mapping.get(component, {})
                    var_name = component_config.get("Variable")
                    norm_vmin, norm_vmax = component_config.get(
                        "ValueRange", norm_range
                    )
                    var_names[i] = var_name
                    norm_ranges[i] = norm_vmin, norm_vmax
        return var_names, norm_ranges

    def get_color_mapping(
        self, ds_id: str, var_name: str
    ) -> tuple[str, str, tuple[float, float]]:

        if split_var_assignment(var_name)[1]:
            # var_name is an assignment expression, the result is unknown yet,
            # so we must return default values here.
            return DEFAULT_CMAP_NAME, DEFAULT_CMAP_NORM, DEFAULT_VALUE_RANGE

        cmap_name = None
        cmap_norm = None
        cmap_vmin, cmap_vmax = None, None
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            color_mapping = color_mappings.get(var_name)
            if color_mapping:
                assert "ColorFile" not in color_mapping
                cmap_name = color_mapping.get("ColorBar")
                cmap_norm = color_mapping.get("Norm")
                cmap_vmin, cmap_vmax = color_mapping.get("ValueRange", (None, None))

        cmap_range = cmap_vmin, cmap_vmax
        if cmap_name and cmap_norm and None not in cmap_range:
            # noinspection PyTypeChecker
            return cmap_name, cmap_norm, cmap_range

        ds = self.get_dataset(ds_id, expected_var_names=[var_name])
        var = ds[var_name]
        valid_range = get_var_valid_range(var)
        return get_var_cmap_params(var, cmap_name, cmap_norm, cmap_range, valid_range)

    def _get_cm_styles(self) -> tuple[dict[str, Any], ColormapRegistry]:
        custom_colormaps = {}
        cm_styles = {}

        # go through CustomColorMaps
        for ctx_cmap in self.config.get("CustomColorMaps", []):
            custom_colormap, config_parse = create_colormap_from_config(ctx_cmap)
            custom_colormaps[custom_colormap.cm_name] = custom_colormap
        # go through Styles
        for style in self.config.get("Styles", []):
            style_id = style["Identifier"]
            color_mappings = dict()
            for var_name, color_mapping in style["ColorMappings"].items():
                if "ColorFile" in color_mapping:
                    custom_cmap_path = self.get_config_path(
                        color_mapping, "ColorMappings", path_entry_name="ColorFile"
                    )
                    custom_colormap = custom_colormaps.get(custom_cmap_path)
                    if custom_colormap is None:
                        custom_colormap = load_custom_colormap(custom_cmap_path)
                        if custom_colormap is not None:
                            custom_colormaps[custom_cmap_path] = custom_colormap
                    if custom_colormap is not None:
                        color_mappings[var_name] = {
                            "ColorBar": custom_colormap.cm_name,
                            "ValueRange": (
                                custom_colormap.norm.vmin,
                                custom_colormap.norm.vmax,
                            ),
                        }
                else:
                    if color_mapping.get("ColorBar") in custom_colormaps:
                        cmap = custom_colormaps[color_mapping["ColorBar"]]
                        if (
                            cmap.cm_type == "categorical"
                            and "ValueRange" in color_mapping
                        ):
                            LOG.warning(
                                f"Custom color map {color_mapping['ColorBar']!r} is "
                                "categorical. ValueRange is ignored."
                            )
                            color_mapping = dict(
                                ColorBar=color_mapping["ColorBar"],
                                ValueRange=[
                                    min(cmap.values),
                                    max(cmap.values),
                                ],
                            )
                        if "ValueRange" not in color_mapping:
                            color_mapping = dict(
                                ColorBar=color_mapping["ColorBar"],
                                ValueRange=[
                                    min(cmap.values),
                                    max(cmap.values),
                                ],
                            )
                    color_mappings[var_name] = dict(color_mapping)

            cm_styles[style_id] = color_mappings

        return cm_styles, ColormapRegistry(*custom_colormaps.values())

    def get_color_mappings(self, ds_id: str) -> Optional[dict[str, dict[str, Any]]]:
        dataset_config = self.get_dataset_config(ds_id)
        style_id = dataset_config.get("Style", "default")
        return self._cm_styles.get(style_id, {})

    def _get_dataset_entry(self, ds_id: str) -> tuple[MultiLevelDataset, ServerConfig]:
        if ds_id not in self._dataset_cache:
            with self.rlock:
                self._set_dataset_entry(self._create_dataset_entry(ds_id))
        return self._dataset_cache[ds_id]

    def _set_dataset_entry(
        self, dataset_entry: tuple[MultiLevelDataset, DatasetConfig]
    ):
        ml_dataset, dataset_config = dataset_entry
        self._dataset_cache[ml_dataset.ds_id] = ml_dataset, dataset_config

    def _create_dataset_entry(
        self, ds_id: str
    ) -> tuple[MultiLevelDataset, DatasetConfig]:
        dataset_config = self.get_dataset_config(ds_id)
        ml_dataset = self._open_ml_dataset(dataset_config)
        return ml_dataset, dataset_config

    def _open_ml_dataset(self, dataset_config: DatasetConfig) -> MultiLevelDataset:
        ds_id: str = dataset_config.get("Identifier")
        store_instance_id = dataset_config.get("StoreInstanceId")
        if store_instance_id:
            data_store_pool = self.get_data_store_pool()
            data_store = data_store_pool.get_store(store_instance_id)
            data_id = dataset_config.get("Path")
            open_params = dict(dataset_config.get("StoreOpenParams") or {})
            open_params_schema = data_store.get_open_data_params_schema(data_id=data_id)

            # Inject cache_size=chunk_cache_capacity, if given and possible
            chunk_cache_capacity = self.get_dataset_chunk_cache_capacity(dataset_config)
            if (
                chunk_cache_capacity
                and "cache_size" not in open_params
                and "cache_size" in open_params_schema.properties
            ):
                open_params["cache_size"] = chunk_cache_capacity

            # Inject data_type="mldataset", if possible
            if (
                "data_type" not in open_params
                and "data_type" in open_params_schema.properties
                and "mldataset" in data_store.get_data_types()
            ):
                open_params["data_type"] = "mldataset"

            with self.measure_time(
                tag=f"Opened dataset {ds_id!r}"
                f" from data store"
                f" {store_instance_id!r}"
            ):
                dataset = data_store.open_data(data_id, **open_params)

            if isinstance(dataset, MultiLevelDataset):
                # Expected, nominal case.
                ml_dataset: MultiLevelDataset = dataset
            else:
                # Fallback. Usually results in poor tile computation performance.
                ml_dataset = BaseMultiLevelDataset(dataset)

            ml_dataset.ds_id = ds_id

        else:
            fs_type = dataset_config.get("FileSystem")
            if fs_type != "memory":
                raise ApiError.InvalidServerConfig(
                    f"Invalid FileSystem {fs_type!r}"
                    f" in dataset configuration"
                    f" {ds_id!r}"
                )

            with self.measure_time(
                tag=f"Opened dataset {ds_id!r}" f" from {fs_type!r}"
            ):
                ml_dataset = _open_ml_dataset_from_python_code(self, dataset_config)

        augmentation = dataset_config.get("Augmentation")
        if augmentation:
            script_path = self.get_config_path(
                augmentation, f"'Augmentation' of dataset configuration {ds_id}"
            )
            input_parameters = augmentation.get("InputParameters", {})
            input_parameters = self.eval_config_value(input_parameters)
            callable_name = augmentation.get("Class")
            is_factory = callable_name is not None
            if not is_factory:
                callable_name = augmentation.get("Function", COMPUTE_VARIABLES)
            ml_dataset = augment_ml_dataset(
                ml_dataset,
                script_path,
                callable_name,
                self.get_ml_dataset,
                self.set_ml_dataset,
                input_parameters=input_parameters,
                is_factory=is_factory,
                exception_type=ApiError.InvalidServerConfig,
            )
        return ml_dataset

    def get_legend_label(self, ds_id: str, var_name: str):
        dataset = self.get_dataset(ds_id)
        if var_name in dataset:
            ds = self.get_dataset(ds_id)
            units = ds[var_name].units
            return units
        raise ApiError.NotFound(f'Variable "{var_name}" not found in dataset "{ds_id}"')

    def get_dataset_place_groups(
        self, ds_id: str, base_url: str, load_features=False
    ) -> list[dict]:
        dataset_config = self.get_dataset_config(ds_id)

        place_group_id_prefix = f"DS-{ds_id}-"

        def predicate(pg_id: str, _: Any):
            return pg_id.startswith(place_group_id_prefix)

        place_groups = self.places_ctx.get_cached_place_groups(predicate=predicate)
        if place_groups:
            # for a given dataset, all place groups cached,
            # so the returned list is complete
            return place_groups

        place_groups = self.places_ctx.load_place_groups(
            dataset_config.get("PlaceGroups", []),
            base_url,
            is_global=False,
            load_features=load_features,
            qualifiers=[ds_id],
        )

        for place_group in place_groups:
            place_group_id = place_group_id_prefix + place_group["id"]
            self.places_ctx.set_cached_place_group(place_group_id, place_group)

        return place_groups

    def get_dataset_place_group(
        self, ds_id: str, place_group_id: str, base_url: str, load_features=False
    ) -> dict:
        place_groups = self.get_dataset_place_groups(
            ds_id, base_url, load_features=False
        )
        for place_group in place_groups:
            if place_group_id == place_group["id"]:
                if load_features:
                    self.places_ctx.load_place_group_features(place_group)
                return place_group
        raise ApiError.NotFound(f'Place group "{place_group_id}" not found')

    def get_dataset_and_coord_variable(self, ds_name: str, dim_name: str):
        ds = self.get_dataset(ds_name)
        if dim_name not in ds.coords:
            raise ApiError.NotFound(
                f"Dimension {dim_name!r} has no coordinates" f" in dataset {ds_name!r}"
            )
        return ds, ds.coords[dim_name]

    @classmethod
    def get_var_indexers(
        cls,
        ds_name: str,
        var_name: str,
        var: xr.DataArray,
        dim_names: list[str],
        query_params: dict[str, str],
    ) -> dict[str, Any]:
        var_indexers = dict()
        for dim_name in dim_names:
            if dim_name not in var.coords:
                raise ApiError.BadRequest(
                    f"Dimension {dim_name!r} of variable {var_name!r}"
                    f" in dataset {ds_name!r} has no coordinates"
                )
            coord_var = var.coords[dim_name]
            dim_value_str = query_params.get(dim_name)
            try:
                if dim_value_str is None:
                    var_indexers[dim_name] = coord_var.values[0]
                elif dim_value_str == "current":
                    var_indexers[dim_name] = coord_var.values[-1]
                elif np.issubdtype(coord_var.dtype, np.floating):
                    var_indexers[dim_name] = float(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.integer):
                    var_indexers[dim_name] = int(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.datetime64):
                    if "/" in dim_value_str:
                        date_str_1, date_str_2 = dim_value_str.split("/", maxsplit=1)
                        var_indexer_1 = pd.to_datetime(date_str_1)
                        var_indexer_2 = pd.to_datetime(date_str_2)
                        var_indexers[dim_name] = (
                            var_indexer_1 + (var_indexer_2 - var_indexer_1) / 2
                        )
                    else:
                        date_str = dim_value_str
                        var_indexers[dim_name] = pd.to_datetime(date_str)
                else:
                    raise ValueError(
                        f"unable to convert value"
                        f" {dim_value_str!r} to {coord_var.dtype!r}"
                    )
            except ValueError as e:
                raise ApiError.BadRequest(
                    f"{dim_value_str!r} is not a valid value"
                    f" for dimension {dim_name!r} "
                    f"of variable {var_name!r} of dataset {ds_name!r}"
                ) from e
        return var_indexers

    def get_dataset_chunk_cache_capacity(
        self, dataset_config: DatasetConfig
    ) -> Optional[int]:
        cache_size = self.get_chunk_cache_capacity(dataset_config, "ChunkCacheSize")
        if cache_size is None:
            cache_size = self.get_chunk_cache_capacity(
                self.config, "DatasetChunkCacheSize"
            )
        return cache_size

    @classmethod
    def get_chunk_cache_capacity(
        cls, config: Mapping[str, Any], cache_size_key: str
    ) -> Optional[int]:
        cache_size = config.get(cache_size_key, None)
        if not cache_size:
            return None
        elif isinstance(cache_size, str):
            try:
                cache_size = parse_mem_size(cache_size)
            except ValueError:
                raise ApiError.InvalidServerConfig(f"Invalid {cache_size_key}")
        elif not isinstance(cache_size, int) or cache_size < 0:
            raise ApiError.InvalidServerConfig(f"Invalid {cache_size_key}")
        return cache_size


def _open_ml_dataset_from_python_code(
    ctx: DatasetsContext, dataset_config: DatasetConfig
) -> MultiLevelDataset:
    ds_id = dataset_config.get("Identifier")
    path = ctx.get_config_path(dataset_config, f"dataset configuration {ds_id}")
    callable_name = dataset_config.get("Class")
    is_factory = callable_name is not None
    if not is_factory:
        callable_name = dataset_config.get("Function", COMPUTE_DATASET)
    input_dataset_ids = dataset_config.get("InputDatasets", [])
    input_parameters = dataset_config.get("InputParameters", {})
    input_parameters = ctx.eval_config_value(input_parameters)
    chunk_cache_capacity = ctx.get_dataset_chunk_cache_capacity(dataset_config)
    if chunk_cache_capacity:
        warnings.warn(
            "chunk cache size is not effective for" " datasets computed from scripts"
        )
    for input_dataset_id in input_dataset_ids:
        if not ctx.get_dataset_config(input_dataset_id):
            raise ApiError.InvalidServerConfig(
                f"Invalid dataset configuration {ds_id!r}:"
                f" Input dataset {input_dataset_id!r} of"
                f" callable {callable_name!r}"
                f" must reference another dataset"
            )
    return open_ml_dataset_from_python_code(
        path,
        callable_name=callable_name,
        input_ml_dataset_ids=input_dataset_ids,
        input_ml_dataset_getter=ctx.get_ml_dataset,
        input_parameters=input_parameters,
        is_factory=is_factory,
        ds_id=ds_id,
        exception_type=ApiError.InvalidServerConfig,
    )


def _is_not_empty(prefix):
    return prefix != "" and prefix != "/" and prefix != "\\"


def _get_roots(paths):
    list.sort(paths)
    prefixes = _get_common_prefixes(paths)
    if len(prefixes) < 1:
        roots = [""]
    else:
        # perform further step to merge prefixes with same start
        prefixes = list(set(prefixes))
        prefixes.sort()
        roots = []
        root_candidate = prefixes[0]
        for root in prefixes[1:]:
            common_root = os.path.commonprefix([root_candidate, root])
            if _is_not_empty(common_root):
                root_candidate = common_root
            else:
                roots.append(root_candidate)
                root_candidate = root
        roots.append(root_candidate)
    if roots[0] == "":
        roots.append(roots.pop(0))
    return roots


def _get_common_prefixes(p):
    # Recursively examine a list of paths for common prefixes:
    # If no common prefix is found, split the list in half and
    # examine each half separately
    prefix = os.path.commonprefix(p)
    # ensure prefix does not end with full or partial directory
    # or file name
    prefix = prefix[: max(_lastindex(prefix, "/"), _lastindex(prefix, "\\"), 0)]
    if _is_not_empty(prefix) or len(p) == 1:
        return [prefix]
    else:
        return _get_common_prefixes(p[: int(len(p) / 2)]) + _get_common_prefixes(
            p[int(len(p) / 2) :]
        )


def _get_selected_dataset_config(
    store_dataset_id: str, store_instance_id: str, dataset_config_base: dict
) -> dict:
    LOG.debug(f"Selected dataset {store_dataset_id!r}")
    dataset_config = dict(StoreInstanceId=store_instance_id, **dataset_config_base)
    if "Identifier" in dataset_config and dataset_config["Path"] != store_dataset_id:
        raise ApiError.InvalidServerConfig(
            "User-defined identifiers can only be assigned"
            " to datasets with non-wildcard paths."
        )
    elif "Identifier" not in dataset_config:
        dataset_config["Path"] = store_dataset_id
        dataset_config["Identifier"] = (
            f"{store_instance_id}{STORE_DS_ID_SEPARATOR}{store_dataset_id}"
        )
    return dataset_config


def _lastindex(prefix, symbol):
    try:
        return prefix.rindex(symbol)
    except ValueError:
        return -1


_MULTI_LEVEL_DATASET_OPENERS = {
    "memory": _open_ml_dataset_from_python_code,
}


def _is_wildard(string: str) -> bool:
    return "?" in string or "*" in string



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from functools import cached_property
from typing import Optional
from collections.abc import Mapping

import fsspec

from xcube.webapi.common.context import ResourcesContext


class ViewerContext(ResourcesContext):
    """Context for xcube Viewer API."""

    @cached_property
    def config_items(self) -> Optional[Mapping[str, bytes]]:
        if self.config_path is None:
            return None
        return fsspec.get_mapper(self.config_path)

    @cached_property
    def config_path(self) -> Optional[str]:
        if "Viewer" not in self.config:
            return None
        return self.get_config_path(
            self.config["Viewer"].get("Configuration", {}),
            "'Configuration' item of 'Viewer'",
        )



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import abstractmethod, ABC
from typing import Any, List, Optional

import xarray as xr

from xcube.constants import EXTENSION_POINT_DATA_OPENERS
from xcube.constants import EXTENSION_POINT_DATA_WRITERS
from xcube.util.assertions import assert_given
from xcube.util.extension import Extension
from xcube.util.extension import ExtensionPredicate
from xcube.util.extension import ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry
from .datatype import DataType
from .datatype import DataTypeLike
from .error import DataStoreError


#######################################################
# Data accessor instantiation and registry query
#######################################################


def new_data_opener(
    opener_id: str,
    extension_registry: Optional[ExtensionRegistry] = None,
    **opener_params,
) -> "DataOpener":
    """Get an instance of the data opener identified by *opener_id*.

    The optional, extra opener parameters *opener_params* may
    be used by data store (``xcube.core.store.DataStore``)
    implementations so they can share their internal state with the opener.

    Args:
        opener_id: The data opener identifier.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.
        **opener_params: Extra opener parameters.

    Returns:
        A data opener instance.
    """
    assert_given(opener_id, "opener_id")
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_OPENERS, opener_id):
        raise DataStoreError(f"A data opener named" f" {opener_id!r} is not registered")
    return extension_registry.get_component(EXTENSION_POINT_DATA_OPENERS, opener_id)(
        **opener_params
    )


def new_data_writer(
    writer_id: str,
    extension_registry: Optional[ExtensionRegistry] = None,
    **writer_params,
) -> "DataWriter":
    """Get an instance of the data writer identified by *writer_id*.

    The optional, extra writer parameters *writer_params* may be used by
    data store (``xcube.core.store.DataStore``) implementations so they
    can share their internal state with the writer.

    Args:
        writer_id: The data writer identifier.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.
        **writer_params: Extra writer parameters.

    Returns:
        A data writer instance.
    """
    assert_given(writer_id, "writer_id")
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_WRITERS, writer_id):
        raise DataStoreError(f"A data writer named" f" {writer_id!r} is not registered")
    return extension_registry.get_component(EXTENSION_POINT_DATA_WRITERS, writer_id)(
        **writer_params
    )


def find_data_opener_extensions(
    predicate: ExtensionPredicate = None,
    extension_registry: Optional[ExtensionRegistry] = None,
) -> list[Extension]:
    """Get registered data opener extensions using the optional
    filter function *predicate*.

    Args:
        predicate: An optional filter function.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.

    Returns:
        List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_OPENERS, predicate=predicate
    )


def find_data_writer_extensions(
    predicate: ExtensionPredicate = None,
    extension_registry: Optional[ExtensionRegistry] = None,
) -> list[Extension]:
    """Get registered data writer extensions using the optional filter
    function *predicate*.

    Args:
        predicate: An optional filter function.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.

    Returns:
        List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_WRITERS, predicate=predicate
    )


def get_data_accessor_predicate(
    data_type: DataTypeLike = None, format_id: str = None, storage_id: str = None
) -> ExtensionPredicate:
    """Get a predicate that checks if a data accessor extensions's name is
    compliant with *data_type*, *format_id*, *storage_id*.

    Args:
        data_type: Optional data data type to be supported. May be given
            as type alias name, as a type, or as a DataType instance.
        format_id: Optional data format identifier to be supported.
        storage_id: Optional data storage identifier to be supported.

    Returns:
        A filter function.

    Raises:
        DataStoreError: If an error occurs.
    """
    if any((data_type, format_id, storage_id)):
        data_type = DataType.normalize(data_type) if data_type is not None else None

        def _predicate(extension: Extension) -> bool:
            extension_parts = extension.name.split(":", maxsplit=4)
            if storage_id is not None:
                ext_storage_id = extension_parts[2]
                if ext_storage_id != "*" and ext_storage_id != storage_id:
                    return False
            if format_id is not None:
                ext_format_id = extension_parts[1]
                if ext_format_id != "*" and ext_format_id != format_id:
                    return False
            if data_type is not None:
                ext_data_type = DataType.normalize(extension_parts[0])
                if not data_type.is_super_type_of(ext_data_type):
                    return False
            return True

    else:
        # noinspection PyUnusedLocal
        def _predicate(extension: Extension) -> bool:
            return True

    return _predicate


#######################################################
# Classes
#######################################################


class DataOpener(ABC):
    """An interface that specifies a parameterized `open_data()` operation.

    Possible open parameters are implementation-specific and
    are described by a JSON Schema.

    Note this interface uses the term "opener" to underline the expected
    laziness of the operation. For example, when a xarray.Dataset is
    returned from a Zarr directory, the actual data is represented by
    Dask arrays and will be loaded only on-demand.
    """

    @abstractmethod
    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *open_params* to
        :meth:`open_data`.
        If *data_id* is given, the returned schema will be tailored
        to the constraints implied by the identified data resource.
        Some openers might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will be
        less restrictive.

        Args:
            data_id: An optional data resource identifier.

        Returns:
            The schema for the parameters in *open_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def open_data(self, data_id: str, **open_params) -> Any:
        """Open the data resource given by the data resource identifier
        *data_id* using the supplied *open_params*.

        Raises if *data_id* does not exist.

        Args:
            data_id: The data resource identifier.
            **open_params: Opener-specific parameters.

        Returns:
            An xarray.Dataset instance.

        Raises:
            DataStoreError: If an error occurs.
        """


class DataDeleter(ABC):
    """An interface that specifies a parameterized `delete_data()` operation.

    Possible delete parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_delete_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *delete_params*
        to :meth:`delete_data`.
        If *data_id* is given, the returned schema will be tailored to
        the constraints implied by the identified data resource.
        Some deleters might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will
        be less restrictive.

        Args:
            data_id: An optional data resource identifier.

        Returns:
            The schema for the parameters in *delete_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def delete_data(self, data_id: str, **delete_params):
        """Delete a data resource. Raises if *data_id* does not exist.

        Args:
            data_id: A data resource identifier known to exist.
            **delete_params: Deleter-specific parameters.

        Raises:
            DataStoreError: If an error occurs.
        """


class DataWriter(DataDeleter, ABC):
    """An interface that specifies a parameterized `write_data()` operation.

    Possible write parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_write_data_params_schema(self) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *write_params* to
        :meth:`write_data`.

        Returns:
            The schema for the parameters in *write_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def write_data(
        self, data: Any, data_id: str, replace: bool = False, **write_params
    ) -> str:
        """Write a data resource using the supplied *data_id* and *write_params*.

        Args:
            data: The data resource's in-memory representation to be
                written.
            data_id: A unique data resource identifier.
            replace: Whether to replace an existing data resource.
            **write_params: Writer-specific parameters.

        Returns:
            The data resource identifier used to write the data
            resource.

        Raises:
            DataStoreError: If an error occurs.
        """


class DataTimeSliceUpdater(DataWriter, ABC):
    """An interface that specifies writing of time slice data."""

    @abstractmethod
    def append_data_time_slice(self, data_id: str, time_slice: xr.Dataset):
        """Append a time slice to the identified data resource.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def insert_data_time_slice(self, data_id: str, time_slice: Any, time_index: int):
        """Insert a time slice into the identified data resource at given index.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.
            time_index: The time index.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def replace_data_time_slice(self, data_id: str, time_slice: Any, time_index: int):
        """Replace a time slice in the identified data resource at given index.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.
            time_index: The time index.

        Raises:
            DataStoreError: If an error occurs.
        """



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .accessor import DataDeleter
from .accessor import DataOpener
from .accessor import DataTimeSliceUpdater
from .accessor import DataWriter
from .accessor import find_data_opener_extensions
from .accessor import find_data_writer_extensions
from .accessor import get_data_accessor_predicate
from .accessor import new_data_opener
from .accessor import new_data_writer
from .assertions import assert_valid_config
from .assertions import assert_valid_params
from .datatype import ANY_TYPE
from .datatype import DATASET_TYPE
from .datatype import DataType
from .datatype import DataTypeLike
from .datatype import GEO_DATA_FRAME_TYPE
from .datatype import MULTI_LEVEL_DATASET_TYPE
from .descriptor import DataDescriptor
from .descriptor import DatasetDescriptor
from .descriptor import GeoDataFrameDescriptor
from .descriptor import MultiLevelDatasetDescriptor
from .descriptor import VariableDescriptor
from .descriptor import new_data_descriptor
from .error import DataStoreError
from .fs.registry import new_fs_data_store
from .search import DataSearcher
from .search import DefaultSearchMixin
from .store import DataStore
from .store import MutableDataStore
from .store import find_data_store_extensions
from .store import list_data_store_ids
from .store import get_data_store_class
from .store import get_data_store_params_schema
from .store import new_data_store
from .storepool import DataStoreConfig
from .storepool import DataStoreInstance
from .storepool import DataStorePool
from .storepool import DataStorePoolLike
from .storepool import get_data_store_instance



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import importlib.resources
import os
import pathlib
from typing import Union, Optional

from xcube.constants import LOG
from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import ViewerContext

ENV_VAR_XCUBE_VIEWER_PATH = "XCUBE_VIEWER_PATH"


@api.route("/viewer/config/{*path}")
class ViewerConfigHandler(ApiHandler[ViewerContext]):
    @api.operation(
        operationId="getViewerConfigurationItem",
        summary="Get a configuration item for the xcube viewer.",
    )
    def get(self, path: Optional[str]):
        config_items = self.ctx.config_items
        if config_items is None:
            raise ApiError.NotFound(f"xcube viewer has" f" not been been configured")
        try:
            data = config_items[path]
        except KeyError:
            raise ApiError.NotFound(
                f"The item {path!r} was not found" f" in viewer configuration"
            )
        content_type = self.get_content_type(path)
        self.response.set_header("Content-Length", str(len(data)))
        if content_type is not None:
            self.response.set_header("Content-Type", content_type)
        self.response.write(data)

    @staticmethod
    def get_content_type(path: str) -> Optional[str]:
        filename_ext = path.split("/")[-1].split(".")[-1]
        if filename_ext in ("json", "xml"):
            return f"application/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("csv", "html", "css", "js"):
            return f"text/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("geojson",):
            return "application/geo+json; charset=UTF-8"
        elif filename_ext in ("gif", "png"):
            return f"image/{filename_ext}"
        elif filename_ext in ("jpeg", "jpg"):
            return "image/jpeg"
        return None


_responses = {
    200: {
        "description": "OK",
        "content": {
            "text/html": {
                "schema": {"type": "string"},
            },
        },
    }
}

_viewer_module = "xcube.webapi.viewer"
_data_dir = "dist"
_default_filename = "index.html"


@api.static_route(
    "/viewer",
    default_filename=_default_filename,
    summary="Brings up the xcube Viewer webpage",
    responses=_responses,
)
def get_local_viewer_path() -> Union[None, str, pathlib.Path]:
    local_viewer_path = os.environ.get(ENV_VAR_XCUBE_VIEWER_PATH)
    if local_viewer_path:
        return local_viewer_path
    try:
        with importlib.resources.files(_viewer_module) as path:
            local_viewer_path = path / _data_dir
            if (local_viewer_path / _default_filename).is_file():
                return local_viewer_path
    except ImportError:
        pass
    LOG.warning(
        f"Cannot find {_data_dir}/{_default_filename}"
        f" in {_viewer_module}, consider setting environment variable"
        f" {ENV_VAR_XCUBE_VIEWER_PATH}",
        exc_info=True,
    )
    return None



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .base import CRS_CRS84
from .base import CRS_WGS84
from .base import DEFAULT_TOLERANCE
from .base import GridMapping



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
import shutil
import warnings
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from collections.abc import Iterable, Mapping

import botocore.exceptions
import pandas as pd
import s3fs
import urllib3.util
import xarray as xr
import zarr
from deprecated import deprecated

from xcube.constants import EXTENSION_POINT_DATASET_IOS
from xcube.constants import (
    FORMAT_NAME_CSV,
    FORMAT_NAME_MEM,
    FORMAT_NAME_NETCDF4,
    FORMAT_NAME_ZARR,
)
from xcube.core.timeslice import (
    append_time_slice,
    insert_time_slice,
    replace_time_slice,
)
from xcube.core.verify import assert_cube
from xcube.util.plugin import ExtensionComponent, get_extension_registry

_DEPRECATION_REASON = "Functionality is redundant. Use xcube.core.store API instead."
_DEPRECATION_VERSION = "0.12.1"


# Note, we cannot remove this deprecated code as long as
# xcube.core.xarray.DatasetAccessor.open() is using it.
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def open_cube(input_path: str, format_name: str = None, **kwargs) -> xr.Dataset:
    """Open a xcube dataset from *input_path*.
    If *format* is not provided it will be guessed from *input_path*.

    Args:
        input_path: input path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments

    Returns:
        xcube dataset
    """
    return open_dataset(input_path, format_name=format_name, is_cube=True, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def write_cube(
    cube: xr.Dataset,
    output_path: str,
    format_name: str = None,
    cube_asserted: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Write a xcube dataset to *output_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        cube: xcube dataset to be written.
        output_path: output path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        xcube dataset *cube*
    """
    if not cube_asserted:
        assert_cube(cube)
    return write_dataset(cube, output_path, format_name=format_name, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def open_dataset(
    input_path: str, format_name: str = None, is_cube: bool = False, **kwargs
) -> xr.Dataset:
    """Open a dataset from *input_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        input_path: input path
        format_name: format, e.g. "zarr" or "netcdf4"
        is_cube: Whether a ValueError will be raised, if the dataset
            read from *input_path* is not a xcube dataset.
        **kwargs: format-specific keyword arguments

    Returns:
        dataset object
    """
    format_name = format_name if format_name else guess_dataset_format(input_path)
    if format_name is None:
        raise ValueError("Unknown input format")
    dataset_io = find_dataset_io(format_name, modes=["r"])
    if dataset_io is None:
        raise ValueError(f"Unknown input format {format_name!r} for {input_path}")
    dataset = dataset_io.read(input_path, **kwargs)
    if is_cube:
        assert_cube(dataset)
    return dataset


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def write_dataset(
    dataset: xr.Dataset, output_path: str, format_name: str = None, **kwargs
) -> xr.Dataset:
    """Write dataset to *output_path*.
    If *format* is not provided it will be guessed from *output_path*.

    Args:
        dataset: Dataset to be written.
        output_path: output path
        format_name: format, e.g. "zarr" or "netcdf4"
        **kwargs: format-specific keyword arguments

    Returns:
        the input dataset
    """
    format_name = format_name if format_name else guess_dataset_format(output_path)
    if format_name is None:
        raise ValueError("Unknown output format")
    dataset_io = find_dataset_io(format_name, modes=["w"])
    if dataset_io is None:
        raise ValueError(f"Unknown output format {format_name!r} for {output_path}")
    dataset_io.write(dataset, output_path, **kwargs)
    return dataset


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class DatasetIO(ExtensionComponent, metaclass=ABCMeta):
    """An abstract base class that represents dataset input/output.

    Args:
        name: A unique dataset I/O identifier.
    """

    def __init__(self, name: str):
        super().__init__(EXTENSION_POINT_DATASET_IOS, name)

    @property
    def description(self) -> str:
        """Returns:
        A description for this input processor
        """
        return self.get_metadata_attr("description", "")

    @property
    def ext(self) -> str:
        """The primary filename extension used by this dataset I/O."""
        return self.get_metadata_attr("ext", "")

    @property
    def modes(self) -> set[str]:
        """A set describing the modes of this dataset I/O.
        Must be one or more of "r" (read), "w" (write), and "a" (append).
        """
        return self.get_metadata_attr("modes", set())

    @abstractmethod
    def fitness(self, path: str, path_type: str = None) -> float:
        """Compute a fitness of this dataset I/O in the interval [0 to 1]
        for reading/writing from/to the given *path*.

        Args:
            path: The path or URL.
            path_type: Either "file", "dir", "url", or None.

        Returns:
            the chance in range [0 to 1]
        """
        return 0.0

    def read(self, input_path: str, **kwargs) -> xr.Dataset:
        """Read a dataset from *input_path* using format-specific read parameters *kwargs*."""
        raise NotImplementedError()

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        """ "Write *dataset* to *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        """ "Append *dataset* to existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def insert(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        """ "Insert *dataset* at *index* into existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def replace(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        """ "Replace *dataset* at *index* in existing *output_path* using format-specific write parameters *kwargs*."""
        raise NotImplementedError()

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        """ "Update *dataset* at *output_path* using format-specific open parameters *kwargs*."""
        raise NotImplementedError()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def get_extension(name: str):
    return get_extension_registry().get_extension(EXTENSION_POINT_DATASET_IOS, name)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def find_dataset_io_by_name(name: str):
    extension = get_extension(name)
    if not extension:
        return None
    return extension.component


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def find_dataset_io(
    format_name: str, modes: Iterable[str] = None, default: DatasetIO = None
) -> Optional[DatasetIO]:
    modes = set(modes) if modes else None
    format_name = format_name.lower()
    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)
    for dataset_io in dataset_ios:
        # noinspection PyUnresolvedReferences
        if format_name == dataset_io.name.lower():
            # noinspection PyTypeChecker
            if not modes or modes.issubset(dataset_io.modes):
                return dataset_io
    for dataset_io in dataset_ios:
        # noinspection PyUnresolvedReferences
        if format_name == dataset_io.ext.lower():
            # noinspection PyTypeChecker
            if not modes or modes.issubset(dataset_io.modes):
                return dataset_io
    return default


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def guess_dataset_format(path: str) -> Optional[str]:
    """Guess a dataset format for a file system path or URL given by *path*.

    Args:
        path: A file system path or URL.

    Returns:
        The name of a dataset format guessed from *path*.
    """
    dataset_io_fitness_list = guess_dataset_ios(path)
    if dataset_io_fitness_list:
        return dataset_io_fitness_list[0][0].name
    return None


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def guess_dataset_ios(path: str) -> list[tuple[DatasetIO, float]]:
    """Guess suitable DatasetIO objects for a file system path or URL given by *path*.
    Returns a list of (DatasetIO, fitness) tuples, sorted by descending fitness values.
    Fitness values are in the interval (0, 1].
    The first entry is the most appropriate DatasetIO object.

    Args:
        path: A file system path or URL.

    Returns:
        A list of (DatasetIO, fitness) tuples.
    """
    if os.path.isfile(path):
        input_type = "file"
    elif os.path.isdir(path):
        input_type = "dir"
    elif path.find("://") > 0:
        input_type = "url"
    else:
        input_type = None

    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)

    dataset_io_fitness_list = []
    for dataset_io in dataset_ios:
        fitness = dataset_io.fitness(path, path_type=input_type)
        if fitness > 0.0:
            dataset_io_fitness_list.append((dataset_io, fitness))

    dataset_io_fitness_list.sort(key=lambda item: -item[1])
    return dataset_io_fitness_list


def _get_ext(path: str) -> Optional[str]:
    _, ext = os.path.splitext(path)
    return ext.lower()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def query_dataset_io(filter_fn: Callable[[DatasetIO], bool] = None) -> list[DatasetIO]:
    dataset_ios = get_extension_registry().find_components(EXTENSION_POINT_DATASET_IOS)
    if filter_fn is None:
        return dataset_ios
    return list(filter(filter_fn, dataset_ios))


# noinspection PyAbstractClass
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class MemDatasetIO(DatasetIO):
    """An in-memory  dataset I/O. Keeps all datasets in a dictionary.

    Args:
        datasets: The initial datasets as a path to dataset mapping.
    """

    def __init__(self, datasets: dict[str, xr.Dataset] = None):
        super().__init__(FORMAT_NAME_MEM)
        self._datasets = datasets or {}

    @property
    def datasets(self) -> dict[str, xr.Dataset]:
        return self._datasets

    def fitness(self, path: str, path_type: str = None) -> float:
        if path in self._datasets:
            return 1.0
        ext_value = _get_ext(path) == ".mem"
        type_value = 0.0
        return (3 * ext_value + type_value) / 4

    def read(self, path: str, **kwargs) -> xr.Dataset:
        if path in self._datasets:
            return self._datasets[path]
        raise FileNotFoundError(path)

    def write(self, dataset: xr.Dataset, path: str, **kwargs):
        self._datasets[path] = dataset

    def append(self, dataset: xr.Dataset, path: str, **kwargs):
        if path in self._datasets:
            old_ds = self._datasets[path]
            # noinspection PyTypeChecker
            self._datasets[path] = xr.concat(
                [old_ds, dataset],
                dim="time",
                data_vars="minimal",
                coords="minimal",
                compat="equals",
            )
        else:
            self._datasets[path] = dataset.copy()

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            ds = self._datasets[output_path]
            ds.attrs.update(global_attrs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class Netcdf4DatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to NetCDF files."""

    def __init__(self):
        super().__init__(FORMAT_NAME_NETCDF4)

    def fitness(self, path: str, path_type: str = None) -> float:
        ext = _get_ext(path)
        ext_value = ext in {".nc", ".hdf", ".h5"}
        type_value = 0.0
        if path_type == "file":
            type_value = 1.0
        elif path_type is None:
            type_value = 0.5
        else:
            ext_value = 0.0
        return (3 * ext_value + type_value) / 4

    def read(self, input_path: str, **kwargs) -> xr.Dataset:
        return xr.open_dataset(input_path, **kwargs)

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        dataset.to_netcdf(output_path)

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        import os

        temp_path = output_path + ".temp.nc"
        os.rename(output_path, temp_path)
        old_ds = xr.open_dataset(temp_path)
        new_ds = xr.concat(
            [old_ds, dataset],
            dim="time",
            data_vars="minimal",
            coords="minimal",
            compat="equals",
        )
        # noinspection PyUnresolvedReferences
        new_ds.to_netcdf(output_path)
        old_ds.close()
        rimraf(temp_path)

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            import netCDF4

            ds = netCDF4.Dataset(output_path, "r+")
            ds.setncatts(global_attrs)
            ds.close()


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class ZarrDatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to Zarr directories or archives."""

    def __init__(self):
        super().__init__(FORMAT_NAME_ZARR)

    def fitness(self, path: str, path_type: str = None) -> float:
        ext = _get_ext(path)
        ext_value = 0.0
        type_value = 0.0
        if ext == ".zarr":
            ext_value = 1.0
            if path_type == "dir":
                type_value = 1.0
            elif path_type == "url" or path_type is None:
                type_value = 0.5
            else:
                ext_value = 0.0
        else:
            lower_path = path.lower()
            if lower_path.endswith(".zarr.zip"):
                ext_value = 1.0
                if path_type == "file":
                    type_value = 1.0
                elif path_type is None:
                    type_value = 0.5
                else:
                    ext_value = 0.0
            else:
                if path_type == "dir":
                    type_value = 1.0
                elif path_type == "url":
                    type_value = 0.5
        return (3 * ext_value + type_value) / 4

    def read(
        self,
        path: str,
        s3_kwargs: dict[str, Any] = None,
        s3_client_kwargs: dict[str, Any] = None,
        max_cache_size: int = None,
        **kwargs,
    ) -> xr.Dataset:
        """Read dataset from some Zarr storage.

        Args:
            path: File path or object storage URL.
            s3_kwargs: if *path* is an object storage URL, keyword-
                arguments passed to S3 file system, that is
                ``s3fs.S3FileSystem(**s3_kwargs, ...)``.
            s3_client_kwargs: if *path* is an object storage URL,
                keyword-arguments passed to S3 (boto3) client, that is
                ``s3fs.S3FileSystem(...,
                client_kwargs=s3_client_kwargs)``.
            max_cache_size: if this is a positive integer, the store
                will be wrapped in an in-memory cache, that is ``store =
                zarr.LRUStoreCache(store, max_size=max_cache_size)``.
            **kwargs: Keyword-arguments passed to xarray Zarr adapter,
                that is ``xarray.open_zarr(..., **kwargs)``. In
                addition, the parameter **

        Returns:

        """
        path_or_store = path
        consolidated = False
        if isinstance(path, str):
            path_or_store, consolidated = get_path_or_s3_store(
                path_or_store,
                s3_kwargs=s3_kwargs,
                s3_client_kwargs=s3_client_kwargs,
                mode="r",
            )
            if max_cache_size is not None and max_cache_size > 0:
                path_or_store = zarr.LRUStoreCache(
                    path_or_store, max_size=max_cache_size
                )
        return xr.open_zarr(path_or_store, consolidated=consolidated, **kwargs)

    def write(
        self,
        dataset: xr.Dataset,
        output_path: str,
        compressor: dict[str, Any] = None,
        chunksizes: dict[str, int] = None,
        packing: dict[str, dict[str, Any]] = None,
        s3_kwargs: dict[str, Any] = None,
        s3_client_kwargs: dict[str, Any] = None,
        **kwargs,
    ):
        path_or_store, _ = get_path_or_s3_store(
            output_path,
            s3_kwargs=s3_kwargs,
            s3_client_kwargs=s3_client_kwargs,
            mode="w",
        )
        encoding = self._get_write_encodings(dataset, compressor, chunksizes, packing)
        consolidated = kwargs.pop("consolidated", True)
        dataset.to_zarr(
            path_or_store,
            mode="w",
            encoding=encoding,
            consolidated=consolidated,
            **kwargs,
        )

    @classmethod
    def _get_write_encodings(cls, dataset, compressor, chunksizes, packing):
        encoding = None
        if chunksizes:
            encoding = {}
            for var_name in dataset.data_vars:
                var = dataset[var_name]
                chunks: list[int] = []
                for i in range(len(var.dims)):
                    dim_name = var.dims[i]
                    if dim_name in chunksizes:
                        chunks.append(chunksizes[dim_name])
                    else:
                        chunks.append(var.shape[i])
                encoding[var_name] = dict(chunks=chunks)
        if packing:
            if encoding:
                for var_name in packing.keys():
                    if var_name in encoding.keys():
                        encoding[var_name].update(dict(packing[var_name]))
                    else:
                        encoding[var_name] = dict(packing[var_name])
            else:
                encoding = {}
                for var_name in packing.keys():
                    encoding[var_name] = dict(packing[var_name])

        if compressor:
            compressor = zarr.Blosc(**compressor)

            if encoding:
                for var_name in encoding.keys():
                    encoding[var_name].update(compressor=compressor)
            else:
                encoding = {
                    var_name: dict(compressor=compressor)
                    for var_name in dataset.data_vars
                }
        return encoding

    def append(self, dataset: xr.Dataset, output_path: str, **kwargs):
        append_time_slice(output_path, dataset)

    def insert(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        insert_time_slice(output_path, index, dataset)

    def replace(self, dataset: xr.Dataset, index: int, output_path: str, **kwargs):
        replace_time_slice(output_path, index, dataset)

    def update(self, output_path: str, global_attrs: dict[str, Any] = None, **kwargs):
        if global_attrs:
            import zarr

            ds = zarr.open_group(output_path, mode="r+", **kwargs)
            ds.attrs.update(global_attrs)
            zarr.consolidate_metadata(output_path)


# noinspection PyAbstractClass
@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
class CsvDatasetIO(DatasetIO):
    """A dataset I/O that reads from / writes to CSV files."""

    def __init__(self):
        super().__init__(FORMAT_NAME_CSV)

    def fitness(self, path: str, path_type: str = None) -> float:
        if path_type == "dir":
            return 0.0
        ext = _get_ext(path)
        ext_value = {".csv": 1.0, ".txt": 0.5, ".dat": 0.2}.get(ext, 0.1)
        type_value = {"file": 1.0, "url": 0.5, None: 0.5}.get(path_type, 0.0)
        return (3 * ext_value + type_value) / 4

    def read(self, path: str, **kwargs) -> xr.Dataset:
        return xr.Dataset.from_dataframe(pd.read_csv(path, **kwargs))

    def write(self, dataset: xr.Dataset, output_path: str, **kwargs):
        dataset.to_dataframe().to_csv(output_path, **kwargs)


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def rimraf(*paths: str):
    """The UNIX command `rm -rf` for xcube.
    Recursively remove directory or single file.

    Args:
        *paths: one or more directories or files
    """
    for path in paths:
        if os.path.isdir(path):
            try:
                shutil.rmtree(path)
            except OSError:
                warnings.warn(f"failed to remove file {path}")
        elif os.path.isfile(path):
            try:
                os.remove(path)
            except OSError:
                warnings.warn(f"failed to remove file {path}")
                pass


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def get_path_or_s3_store(
    path_or_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    mode: str = "r",
) -> tuple[Union[str, dict], bool]:
    """If *path_or_url* is an object storage URL, return a object storage
    Zarr store (mapping object) using *s3_client_kwargs* and *mode* and
    a flag indicating whether the Zarr datasets is consolidated.

    Otherwise *path_or_url* is interpreted as a local file system path,
    returned as-is plus a flag indicating whether the Zarr datasets
    is consolidated.

    Args:
        path_or_url: A path or a URL.
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        mode: access mode "r" or "w". "r" is default.

    Returns:
        A tuple (path_or_obs_store, consolidated).
    """
    if is_s3_url(path_or_url) or s3_kwargs is not None or s3_client_kwargs is not None:
        s3, root = parse_s3_fs_and_root(
            path_or_url,
            s3_kwargs=s3_kwargs,
            s3_client_kwargs=s3_client_kwargs,
            mode=mode,
        )
        consolidated = mode == "r" and s3.exists(f"{root}/.zmetadata")
        return (
            s3fs.S3Map(root=root, s3=s3, check=False, create=mode == "w"),
            consolidated,
        )
    else:
        consolidated = os.path.exists(os.path.join(path_or_url, ".zmetadata"))
        return path_or_url, consolidated


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def parse_s3_fs_and_root(
    s3_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    mode: str = "r",
) -> tuple[s3fs.S3FileSystem, str]:
    """Parses *s3_url*, *s3_kwargs*, *s3_client_kwargs* and returns a
    new tuple (*obs_fs*, *root_path*). For example::

        obs_fs, root_path = parse_s3_fs_and_root(s3_url, s3_kwargs, s3_client_kwargs)
        obs_map = s3fs.S3Map(root=root_path, s3=obs_fs)

    Args:
        s3_url: Object storage URL, e.g. "s3://bucket/root", or
            "https://bucket.s3.amazonaws.com/root".
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        mode: Access mode "r" or "w",  "r" is default.

    Returns:
        A tuple (*obs_fs*, *root_path*).
    """

    root, s3_kwargs, s3_client_kwargs = parse_s3_url_and_kwargs(
        s3_url, s3_kwargs=s3_kwargs, s3_client_kwargs=s3_client_kwargs
    )
    s3 = new_s3_file_system(
        s3_kwargs=s3_kwargs,
        s3_client_kwargs=s3_client_kwargs,
        check_path=root if mode == "r" else None,
    )
    return s3, root


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def new_s3_file_system(
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
    s3_config_param_name: str = "s3_kwargs",
    check_path: str = None,
) -> s3fs.S3FileSystem:
    """Wrapper for s3fs.S3FileSystem() constructor that issues warnings
    in case the file system can not be created.

    Args:
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.
        s3_config_param_name: the name of the configuration parameter.
        check_path: an optional root path. If given, we call
            s3.exists(check_path) as a check whether S3 file system is
            valid.

    Returns:
        A s3fs.S3FileSystem instance.
    """

    s3_kwargs = s3_kwargs or {}
    if "use_listings_cache" not in s3_kwargs:
        # The default is not to cache any directory listings
        # because we want current contents
        s3_kwargs["use_listings_cache"] = False
    client_kwargs = s3_kwargs.pop("client_kwargs", {})
    client_kwargs.update(s3_client_kwargs or {})
    try:
        s3 = s3fs.S3FileSystem(**s3_kwargs, client_kwargs=client_kwargs)
        if check_path is not None:
            # Force potential NoCredentialsError
            s3.exists(check_path)
        return s3
    except botocore.exceptions.NoCredentialsError:
        if not s3_kwargs.get("anon"):
            warnings.warn(
                "No object storage credentials were"
                " passed or found.\n"
                "If you intend to access a public object storage,"
                " please pass " + s3_config_param_name + '={"anon": True, ...}\n'
            )
        raise


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def parse_s3_url_and_kwargs(
    s3_url: str,
    s3_kwargs: Mapping[str, Any] = None,
    s3_client_kwargs: Mapping[str, Any] = None,
) -> tuple[str, dict[str, Any], dict[str, Any]]:
    """Parses *obs_url*, *s3_kwargs*, *s3_client_kwargs* and returns a
    new tuple (*root*, *s3_kwargs*, *s3_client_kwargs*) with updated kwargs whose elements
    can be passed to the s3fs.S3FileSystem and s3fs.S3Map constructors as follows:::

        obs_fs = s3fs.S3FileSystem(**s3_kwargs, client_kwargs=s3_client_kwargs)
        obs_map = s3fs.S3Map(root=root, s3=obs_fs)

    Args:
        s3_url: Object storage URL, e.g. "s3://bucket/root", or
            "https://bucket.s3.amazonaws.com/root".
        s3_kwargs: keyword arguments for S3 file system.
        s3_client_kwargs: keyword arguments for S3 boto3 client.

    Returns:
        A tuple (root, s3_kwargs, s3_client_kwargs).
    """
    endpoint_url, root = split_s3_url(s3_url)

    new_s3_client_kwargs = dict(s3_client_kwargs) if s3_client_kwargs else dict()
    if endpoint_url:
        new_s3_client_kwargs["endpoint_url"] = endpoint_url

    # The following key + secret kwargs are no longer supported in client_kwargs and are now moved into s3_kwargs
    key = secret = None
    if "provider_access_key_id" in new_s3_client_kwargs:
        key = new_s3_client_kwargs.pop("provider_access_key_id")
    if "aws_access_key_id" in new_s3_client_kwargs:
        key = new_s3_client_kwargs.pop("aws_access_key_id")
    if "provider_secret_access_key" in new_s3_client_kwargs:
        secret = new_s3_client_kwargs.pop("provider_secret_access_key")
    if "aws_secret_access_key" in new_s3_client_kwargs:
        secret = new_s3_client_kwargs.pop("aws_secret_access_key")

    new_s3_kwargs = dict(key=key, secret=secret)
    if s3_kwargs:
        new_s3_kwargs.update(**s3_kwargs)

    return root, new_s3_kwargs, new_s3_client_kwargs


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def split_s3_url(path: str) -> tuple[Optional[str], str]:
    """If *path* is a URL, return tuple (endpoint_url, root), otherwise (None, *path*)"""
    url = urllib3.util.parse_url(path)
    if all((url.scheme, url.host, url.path)) and url.scheme != "s3":
        if url.port is not None:
            endpoint_url = f"{url.scheme}://{url.host}:{url.port}"
        else:
            endpoint_url = f"{url.scheme}://{url.host}"
        root = url.path
        if root.startswith("/"):
            root = root[1:]
        return endpoint_url, root
    return None, path


@deprecated(_DEPRECATION_REASON, version=_DEPRECATION_VERSION)
def is_s3_url(path_or_url: str) -> bool:
    """Test if *path_or_url* is a potential object storage URL.

    Args:
        path_or_url: Path or URL to test.

    Returns:
        True, if so.
    """
    return (
        path_or_url.startswith("https://")
        or path_or_url.startswith("http://")
        or path_or_url.startswith("s3://")
    )
